import json
import re
from typing import Type

from langchain_community.chat_models import ChatVertexAI
from langchain_core.exceptions import OutputParserException
from langchain_core.output_parsers import BaseOutputParser, StrOutputParser
from langchain_core.prompts import ChatPromptTemplate
from pydantic import ValidationError

from voice_stream.types import T


def reduce_schema(schema: dict) -> dict:
    if "title" in schema:
        del schema["title"]
    if "type" in schema:
        del schema["type"]
    for prop in schema["properties"]:
        if "title" in schema["properties"][prop]:
            del schema["properties"][prop]["title"]
    return schema


def pydantic_chain(cls):
    return (
        ChatPromptTemplate.from_messages([("human", "{query}")])
        | ChatVertexAI()
        | StrOutputParser()
    )


# main_chain = (
#         RunnableParallel(
#             message=itemgetter("message"),
#             history=itemgetter("history"),
#             assistant=lambda x: x['assistant'] if 'assistant' in x else "default",
#         )
#         | RunnablePassthrough.assign(assistant=routing_chain)
#         | RunnablePassthrough.assign(
#     # RunnableLambda doesn't work here because it breaks streaming.
#     output=RunnableSwitch(
#         "assistant",
#         branches = {_.name: _.chain for _ in assistants},
#         default_branch = [_.chain for _ in assistants if _.name == 'default'][0]
#     )
# )
# )


PYDANTIC_FORMAT_INSTRUCTIONS = """The output should be formatted as a JSON instance that conforms to the JSON schema below.

As an example, for the schema {{"properties": {{"foo": {{"title": "Foo", "description": "a list of strings", "type": "array", "items": {{"type": "string"}}}}}}, "required": ["foo"]}}
the object {{"foo": ["bar", "baz"]}} is a well-formatted instance of the schema. The object {{"properties": {{"foo": ["bar", "baz"]}}}} is not well-formatted.

Here is the output schema:
```
{schema}
```"""


class PydanticV2OutputParser(BaseOutputParser):
    """Parse an output using a pydantic model."""

    pydantic_object: Type[T]
    """The pydantic model to parse."""

    def parse(self, text: str) -> T:
        try:
            # Greedy search for 1st json candidate.
            match = re.search(
                r"\{.*\}", text.strip(), re.MULTILINE | re.IGNORECASE | re.DOTALL
            )
            json_str = match.group() if match else ""
            json_object = json.loads(json_str, strict=False)
            return self.pydantic_object.parse_obj(json_object)

        except (json.JSONDecodeError, ValidationError) as e:
            name = self.pydantic_object.__name__
            msg = f"Failed to parse {name} from completion {text}. Got: {e}"
            raise OutputParserException(msg, llm_output=text)

    def get_format_instructions(self) -> str:
        schema = self.pydantic_object.schema()

        # Remove extraneous fields.
        reduced_schema = schema
        if "title" in reduced_schema:
            del reduced_schema["title"]
        if "type" in reduced_schema:
            del reduced_schema["type"]
        # Ensure json in context is well-formed with double quotes.
        schema_str = json.dumps(reduced_schema)

        return PYDANTIC_FORMAT_INSTRUCTIONS.format(schema=schema_str)

    @property
    def _type(self) -> str:
        return "pydantic"


NEXT_QUESTION_PROMPT = """
You are trying to create an "ActionItem", which is "A task that to be captured in the to-do list.".  

Here is the schema of the structure you need to fill in:
{
        "properties": {
            "name": {"description": "A short name for the task that will be displayed in lists.", "type": "string"},
            "description": {"description": "A detailed description of the task.", "type": "string"},
            "assignee": {"description": "Email address of the person to whom this task is assigned.", "type": "string"},
            "due_date": {"description": "Date on which the task is due.", "format":"date", "type": "string"},
            "done": {"description": "Whether the task is done or not.", "type":"boolean", "default": False},
        },
        "required": ["name", "description", "assignee", "due_date"],
    }
}

At the start of each message, I'll tell you what the information you already have is.

Ask the user questions, one at a time, until you have all the information.
"""

PARSING_PROMPT = """
You are trying to create an "ActionItem", which is "A task that to be captured in the to-do list.".  

The output should be formatted as a JSON instance that conforms to the JSON schema below.

As an example, for the schema {"properties": {"foo": {"title": "Foo", "description": "a list of strings", "type": "array", "items": {"type": "string"}}}, "required": ["foo"]}
the object {"foo": ["bar", "baz"]} is a well-formatted instance of the schema. The object {"properties": {"foo": ["bar", "baz"]}} is not well-formatted.

Here is the output schema:
{
        "properties": {
            "name": {"description": "A short name for the task that will be displayed in lists.", "type": "string"},
            "description": {"description": "A detailed description of the task.", "type": "string"},
            "assignee": {"description": "Email address of the person to whom this task is assigned.", "type": "string"", "format":"email"},
            "due_date": {"description": "Date on which the task is due.", "format":"date", "type": "string"},
            "done": {"description": "Whether the task is done or not.", "type":"boolean"},
        },
    }
}

Today is Wednesday, 1/10/2024

Extract whatever structured information you can from the user's message.  Be conservative, if you aren't sure, just ignore the field.  Don't make up anything.  Don't include any fields in the output that you can explicitly populate based on the input.
"""



import logging
import os
import uuid

from dotenv import load_dotenv
from fastapi import FastAPI, WebSocket
from fastapi.responses import HTMLResponse
from langchain_community.cache import SQLiteCache
from langchain_core.globals import set_llm_cache

from voice_stream import map_step
from voice_stream.integrations.fastapi_streams import (
    fastapi_websocket_text_source,
    fastapi_websocket_text_sink,
)
from voice_stream.integrations.langchain_streams import langchain_step
from voice_stream.types import load_attribute

logger = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.DEBUG, format="%(asctime)s %(name)s %(levelname)s - %(message)s"
)

set_llm_cache(SQLiteCache(database_path=".langchain.db"))

app = FastAPI()

html = """
<!DOCTYPE html>
<html>
    <head>
        <title>Chat</title>
        <style>
        .message-input {
            margin-bottom: 20px;
            width: 300px; /* Width of the text area */
            height: 100px; /* Height of the text area */
        }        
        </style>
    </head>
    <body>
        <h1>WebSocket Chat</h1>
        <form action="" onsubmit="sendMessage(event)">
            <textarea class="message-input" id="messageText" autocomplete="off"></textarea>
            <button>Send</button>
        </form>
        <ul id='messages'>
        </ul>
        <script>
            var newAI = false;
            function appendMessage(role, text) {
                var messages = document.getElementById('messages')
                var message = document.createElement('li')
                var aiPrefix = document.createElement('strong');
                aiPrefix.textContent = role;
                var content = document.createTextNode(text)
                message.appendChild(aiPrefix)
                message.appendChild(content)
                messages.appendChild(message)            
            };
            var ws = new WebSocket("ws://localhost:8000/ws");
            ws.onmessage = function(event) {
                if (newAI) {
                    newAI = false;
                    appendMessage('AI: ', event.data);
                }
                else {
                    var messages = document.getElementById('messages')
                    messages.lastChild.innerHTML += escapeHTML(event.data)
                }
            };
            function escapeHTML(str) {
                return str.replace(/[&<>"']/g, function(match) {
                    const escape = {
                        '&': '&amp;',
                        '<': '&lt;',
                        '>': '&gt;',
                        '"': '&quot;',
                        "'": '&#39;'
                    };
                    return escape[match];
                });
            }
            function sendMessage(event) {
                var input = document.getElementById("messageText")
                ws.send(input.value)
                appendMessage('Human: ', input.value);
                newAI = true;
                input.value = ''
                event.preventDefault()
            }
        </script>
    </body>
</html>
"""


@app.get("/")
async def get():
    return HTMLResponse(html)


# set_debug(True)


load_dotenv()
chain = load_attribute(os.environ["EXAMPLE_CHAIN"])()


@app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket):
    session_id = uuid.uuid4()
    pipe = fastapi_websocket_text_source(websocket)
    pipe = map_step(
        pipe,
        lambda x: {
            "input": {"message": x},
            "config": {"configurable": {"session_id": session_id}},
        },
    )
    pipe = langchain_step(pipe, chain, input_key="input", config_key="config")
    pipe = map_step(pipe, lambda x: x, ignore_none=True)
    await fastapi_websocket_text_sink(pipe, websocket)



from enum import Enum


class AudioFormatError(Exception):
    pass


class AudioFormat(str, Enum):
    WAV_MULAW_8KHZ = "wav_mulaw_8khz"
    OGG_OPUS = "ogg_opus"
    WEBM_OPUS = "webm_opus"
    MP3 = "mp3"



from typing import Dict

from langchain_core.runnables import RunnableBranch
from langchain_core.runnables.base import RunnableLike

from voice_stream.langchain_extensions.parsers import logger


class RunnableSwitch(RunnableBranch):
    """Like RunnableBranch, but just takes a dict"""

    def __init__(
        self,
        input_key: str,
        branches: Dict[
            str,
            RunnableLike,
        ],
        default_branch: RunnableLike,
    ) -> None:
        def check(y):
            def f(x):
                logger.info(f"Checking {x} == {y}")
                return x[input_key] == y

            return f

        b = [*[(check(k), v) for k, v in branches.items()], default_branch]
        super().__init__(*b)



import asyncio
import importlib
import inspect
import logging
from asyncio import Future
from typing import TypeVar, Tuple, List, Optional, Union, Callable, AsyncIterator

logger = logging.getLogger(__name__)

T = TypeVar("T")
Input = TypeVar("Input", contravariant=True)
Output = TypeVar("Output", covariant=True)
FutureOrObj = Union[T, asyncio.Future[T]]

SourceConvertable = Optional[Union[Callable[[], T], T]]


def to_source(x: SourceConvertable) -> AsyncIterator[T]:
    """Creates a source from a callable or value."""
    if callable(x):
        return x()
    elif x is None:
        logger.info("Empty source")
        from voice_stream.basic_streams import empty_source

        return empty_source()
    else:
        logger.info("Single source")
        from voice_stream.basic_streams import single_source

        return single_source(x)


def to_tuple(obj) -> Tuple:
    if isinstance(obj, Tuple):
        return obj
    if isinstance(obj, List):
        return tuple(obj)
    else:
        return (obj,)


def from_tuple(obj):
    return obj[0] if len(obj) == 1 else obj


def is_async_iterator(obj):
    return inspect.iscoroutinefunction(getattr(obj, "__anext__", None)) and hasattr(
        obj, "__aiter__"
    )


def map_future(f: Future[T], func: Callable[[T], Output]) -> Future[Output]:
    """Returns a future that will be resolved with the result of applying func to the result of f.

    Words are made lowercase and punctuation is removed
    before counting.

    Parameters
    ----------
    input_file : str
        Path to text file.

    Returns
    -------
    collections.Counter
        dict-like object where keys are words and values are counts.

    Examples
    --------
    >>> count_words("text.txt")
    """
    loop = asyncio.get_running_loop()
    ret = loop.create_future()

    def callback(fut: Future[T]) -> None:
        try:
            ret.set_result(func(fut.result()))
        except Exception as e:
            ret.set_exception(e)

    f.add_done_callback(callback)
    return ret


async def resolve_obj_or_future(obj: FutureOrObj[T]) -> T:
    """Returns the result of an object or a future"""
    if isinstance(obj, asyncio.Future):
        return await obj
    return obj


def load_attribute(full_path):
    # Split the path into module and attribute
    parts = full_path.split(".")
    module_path = ".".join(parts[:-1])
    attribute_name = parts[-1]

    try:
        # Import the module
        module = importlib.import_module(module_path)
        # Return the attribute
        return getattr(module, attribute_name)
    except ImportError:
        print(f"Module {module_path} not found.")
    except AttributeError:
        print(f"Attribute {attribute_name} not found in module {module_path}.")



import json
import logging
from typing import AsyncIterator, Union

import asyncstdlib
from fastapi import WebSocket
from starlette.websockets import WebSocketDisconnect

logger = logging.getLogger(__name__)


async def fastapi_websocket_text_source(websocket: WebSocket) -> AsyncIterator[str]:
    """Creates an AsyncIterator for text messages coming from the FastAPI websocket."""
    await websocket.accept()
    try:
        while True:
            yield await websocket.receive_text()
    except WebSocketDisconnect:
        pass


async def fastapi_websocket_bytes_source(websocket: WebSocket) -> AsyncIterator[bytes]:
    """Creates an AsyncIterator for binary data coming from the FastAPI websocket."""
    await websocket.accept()
    try:
        while True:
            yield await websocket.receive_bytes()
    except WebSocketDisconnect:
        pass


async def fastapi_websocket_json_source(websocket: WebSocket) -> AsyncIterator[dict]:
    """Creates an AsyncIterator for json data coming from the FastAPI websocket."""
    await websocket.accept()
    try:
        while True:
            yield await websocket.receive_json()
    except WebSocketDisconnect:
        pass


async def fastapi_websocket_text_sink(
    async_iter: AsyncIterator[Union[str, dict]], websocket: WebSocket
) -> None:
    """Takes an async iterator and sends everything to a FastAPI websocket.  Handles strings or dicts."""
    async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
        async for message in owned_aiter:
            if isinstance(message, dict):
                await websocket.send_json(message)
            else:
                await websocket.send_text(message)


async def fastapi_websocket_bytes_sink(
    async_iter: AsyncIterator[bytes], websocket: WebSocket
) -> None:
    """Takes an async iterator and sends everything to a FastAPI websocket.  Expects bytes."""
    try:
        async with asyncstdlib.scoped_iter(async_iter) as owned_aiter:
            async for message in owned_aiter:
                await websocket.send_bytes(message)
    finally:
        logger.debug("fastapi_websocket_bytes_sink done.")



# # Works with the python speech_recognition package
# import asyncio
# from concurrent.futures import ThreadPoolExecutor
# from typing import AsyncIterator
#
# import speech_recognition as sr
#
# from voice_stream import chunk_bytes_step
#
#
# def python_speech_recognition_step(
#     async_iter: AsyncIterator[bytes],
#     recognition_function: str,
#     sample_rate: int,
#     sample_width: int,
#     chunk_size: int = 4096,
#     **recognition_kwargs: dict,
# ) -> AsyncIterator[str]:
#     # This class does triple duty.  It is:
#     # The async iterator that is returned (__aiter__ and __anext__)
#     # The AudioSource for the recognizer.  (__enter__ and __exit__)
#     # The audio stream for the recognizer.  (read)
#     class _AsyncPythonSpeechRecIter(sr.AudioSource):
#         def __init__(self, input):
#             self.stream = self
#             self.loop = asyncio.get_event_loop()
#             self.iter = input.__aiter__()
#             self.buffer = b''
#             self.CHUNK = chunk_size
#             self.SAMPLE_RATE = sample_rate
#             self.SAMPLE_WIDTH = sample_width
#
#         def __aiter__(self):
#                 return self
#
#         async def __anext__(self):
#             loop = asyncio.get_running_loop()
#             with ThreadPoolExecutor() as pool:
#                 result = await loop.run_in_executor(pool, self.blocking_recognize)
#                 return result
#
#         def read(self, size):
#             while True:
#                 if len(self.buffer) >= size:
#                     ret = self.buffer[:size]
#                     self.buffer = self.buffer[size:]
#                     return ret
#                 else:
#                     try:
#                         # Schedule the async function on the event loop and wait for the result
#                         future = asyncio.run_coroutine_threadsafe(self.iter.__anext__(), self.loop)
#                         self.buffer += future.result()  # This blocks the thread until the coroutine is done
#                     except StopAsyncIteration:
#                         ret = self.buffer
#                         self.buffer = b''
#                         return ret
#
#         def blocking_recognize(self):
#             r = sr.Recognizer()
#             audio = r.listen(self)
#             f = getattr(r, recognition_function)
#             return f(audio, **recognition_kwargs)
#
#     pipe = chunk_bytes_step(async_iter, chunk_size)
#     return _AsyncPythonSpeechRecIter(pipe)
#



import logging
from typing import Sequence, List, Any

from langchain.output_parsers import EnumOutputParser
from langchain_core.exceptions import OutputParserException
from langchain_core.output_parsers import BaseOutputParser

logger = logging.getLogger(__name__)


class RobustEnumOutputParser(EnumOutputParser):
    """The regular LangChain EnumOutputParser is too strict.  It doesn't strip punctuation, so we use this one instead."""

    def parse(self, response: str) -> Any:
        no_punc_response = response.replace(".", "").replace(",", "")
        return super().parse(no_punc_response)


class FixedValuesOutputParser(BaseOutputParser):
    """Parses output into one of the fixed string values."""

    values: Sequence[str]
    """The allowed values."""

    @property
    def _valid_values(self) -> List[str]:
        return list(self.values)

    def parse(self, response: str) -> Any:
        no_punc_response = response.replace(".", "").replace(",", "").strip().lower()
        for key in self.values:
            if no_punc_response == key.lower():
                return key
        raise OutputParserException(
            f"Response '{response}' is not one of the "
            f"expected values: {self._valid_values}"
        )

    def get_format_instructions(self) -> str:
        return f"Select one of the following options: {', '.join(self._valid_values)}"
